;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "dtrt-indent" "1.26"
  "Adapt to foreign indentation offsets."
  '((emacs "28.1"))
  :url "https://github.com/jscheid/dtrt-indent"
  :commit "7c372bec8d84c247e4bd0d5599024d66ee300429"
  :revdesc "7c372bec8d84"
  :keywords '("convenience" "files" "languages" "c")
  :authors '(("Julian Scheid" . "julians37@googlemail.com"))
  :maintainers '(("Reuben Thomas" . "rrt@sc3d.org")))
